/*
 *  sPico - Pico 1
 *
 *  Kim Gybels (kim@infogroep.be)
 *  Toon Verwaest (toon@infogroep.be)
 *
 */

#include "escheme.h"

#include <stdio.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h> /* For memset */

#include "Pico.h"
#include "PicoMai.h"

#define SPRINTFBUFLEN   10240
#define MSGBUFLEN       10240

#define MEMSIZE 10*1024*1024

#define NAMELEN         256

static char sprintfbuf[SPRINTFBUFLEN];

static Scheme_Object *exn_catching_apply, *exn_p, *exn_message;
    
/* An error-position. Will be used by _PICO_MARK_ to
 * save start and stop-position of an errormarking */
typedef struct {
  _POS_TYPE_ start;
  _POS_TYPE_ stop;
} errorcontext;

/* Structure to save all properties of a session. */
typedef struct {
  /* Output will be used to save temporary outputstrings */
  char output[SPRINTFBUFLEN]; /* TODO */
  /* Pointer to scheme-thread using current session */
  Scheme_Thread* thread;
  /* A buffer used for message-passing from the pico-engine */
  char message[MSGBUFLEN];
  /* Buffer for a file to load */
  char filename[NAMELEN];
  /* Errormarking */
  errorcontext errorpos;
} SessionContext;

/* Array of available session-objects */
static SessionContext sessions[_MAX_SES_];

/* Scheme-symbol-pointers */
static Scheme_Object *spico_error,*spico_continue,*spico_accept,*spico_wait,*spico_start;

static char *globalfilebuf = NULL;

/* Strings matching all the errors of pico */
static char * errors[] = {
  "done",
  "abstract grammar violation",
  "argument type conflict",
  "text buffer overflow",
  "control violation",
  "too many names",
  "digit required",
  "duplicate session identifier",
  "expression violation",
  "excess token(s)",
  "invalid argument",
  "table index violation",
  "invalid index",
  "illegal character",
  "infinite value",
  "invalid parameter",
  "storage overflow",
  "not a boolean",
  "not a function",
  "not a table",
  "number too large",
  "negative argument",
  "non-matching argument list",
  "right brace expected",
  "right bracket expected",
  "reentrancy violation",
  "reference expected",
  "range error",
  "right parenthesis expected",
  "invalid size",
  "session not active",
  "stack error",
  "invalid tag",
  "too many sessions",
  "non-terminated text",
  "undefined identifier",
  "user error",
  "zero division"
};

/* Pico prints strings in pieces, so we add the given piece to
 * the output-string, everytime we get one. Adding strings is done
 * by Scheme's string-append. If there is no string yet, we set
 * the given string, converted to Scheme-string as output */
_ESC_TYPE_ _PICO_PRINT_(const _SES_TYPE_ session,
    const _STR_TYPE_ s)
{
  strcat(sessions[session].output,s);
  return _ESCAPE_NO_;
}

/* Everytime pico gives us the opportunity to escape evaluation,
 * we take it. First we let scheme do some magic before we actually
 * go back to the scheme-part */
_ESC_TYPE_ _PICO_ESCAPE_(const _SES_TYPE_ s)
{
  scheme_check_threads();
  return _ESCAPE_YES_;
}

/* Messages received from the pico-engine are stored in session */
_NIL_TYPE_ _PICO_MESSAGE_(const _SES_TYPE_ session,
    const _STR_TYPE_ s)
{
  strcpy(sessions[session].message, s);
}

_ESC_TYPE_ display_in_scheme(const _STR_TYPE_ s)
{
  Scheme_Object* print;
  Scheme_Object* args[1];

  /* Get the display function of the scheme-environment. */
  print = scheme_builtin_value("display");

  /* Use the print-function to print the string */
  args[0] = scheme_make_string(s);
  scheme_apply(print, 1, args);
  return _ESCAPE_NO_;
}


/* File-reader used by the Pico-engine (doesn't use scheme's loader...) */
static void read_file(SessionContext session)
{
  FILE *file;
  int ch;
  char *bufptr;
  const char *name = session.message;
  chdir(SCHEME_STR_VAL(scheme_get_param(scheme_config,MZCONFIG_LOAD_DIRECTORY)));
  file = fopen(name, "r");

  if (file == NULL) {
    sprintf(sprintfbuf, "Cannot open file \"%s\"", name);
    scheme_warning(sprintfbuf);
    globalfilebuf = NULL;
    return;
  }
  errno = 0;
  (void)fseek(file, 0, SEEK_END);
  /* fseek return value puzzling, using errno */
  if (errno) {
    sprintf(sprintfbuf, "Error while reading file \"%s\"", name);
    scheme_warning(sprintfbuf);
    globalfilebuf = NULL;
    return;
  }
  if ((bufptr = globalfilebuf =
        (char *)malloc((size_t)ftell(file)+1)) == NULL) {
    sprintf(sprintfbuf,
        "Cannot allocate %ld bytes to load program text",
        ftell(file)+1);
    scheme_warning(sprintfbuf);
  }
  strncpy(session.filename, name, NAMELEN);
  session.filename[NAMELEN-1] = '\0';

  rewind(file);
  while (!feof(file) && ((ch = getc(file)) != EOF))
    *bufptr++ = (char)ch;
  *bufptr = '\0';

  (void)fclose(file);
}

_ESC_TYPE_ _PICO_DISPLAY_(const _SES_TYPE_ session,
    const _STR_TYPE_ s)
{
  /* A pico_display displays the string with scheme's display */
  display_in_scheme(s);
  return _ESCAPE_NO_;
}

/* Marking an error is done by setting start and stop-position in
 * the sessions errorpos. This errorpos will be used to mark the error
 * when escaped from pico-engine */
_NIL_TYPE_ _PICO_MARK_(const _SES_TYPE_ session,
    const _POS_TYPE_ start,
    const _POS_TYPE_ stop)
{
  sessions[session].errorpos.start = start;
  sessions[session].errorpos.stop = stop;
}

/* Will possibly be used to flash a pico-garbage-collect icon */
_NIL_TYPE_ _PICO_RECLAIM_(const _RCL_TYPE_ r)
{
/*  if (r) {
    display_in_scheme("Garbage collect...");
  } else {
    display_in_scheme(" done.\n");
  } */
}

/* A function which checks aliveness of its session-threads.
 * If a session-thread is dead, its session can be killed, and
 * reused afterwords. This will be the case when 'start'-button is 
 * pressed, or when a window is closed */
void cleanup_dead_threads() {
  _SES_TYPE_ i;
  Scheme_Object* deadthread = scheme_builtin_value("thread-dead?");

  for(i = 1; i < _MAX_SES_; ++i) {
    /* Only check if session isn't already 'dead' */
    if((sessions[i].thread != NULL) && scheme_apply(deadthread,1, (Scheme_Object**)(&(sessions[i].thread))) == scheme_true) {
      sessions[i].thread = NULL;
      _PICO_KILL_(i);
    }
  }
}

/* Get a session-id for the given thread. If thread already has
 * a session, this session will be returned. Otherwise the first
 * not-used session-id will be returned. If no session is available
 * the sessionid will be _MAX_SES_, indicating no session is available */
_SES_TYPE_ get_session_id() {

  _SES_TYPE_ i = _MAX_SES_;

  /* First clean up dead threads, to perhaps release a new session
   * to be used by this thread. */
  cleanup_dead_threads();

  for(i = 1; i < _MAX_SES_; ++i) {
    if(sessions[i].thread == scheme_current_thread) {
      return i;
    }
  }

  for(i = 1; i < _MAX_SES_; ++i) {
    if(sessions[i].thread == NULL) {
      sessions[i].thread = scheme_current_thread;
      _PICO_SESSION_(i);
      return i;
    }
  }

  return _MAX_SES_;
}

/* The following contains the main pico <-> scheme interaction */
Scheme_Object *
pico_eval (int argc, Scheme_Object** argv) {
  _RES_TYPE_ res;
  Scheme_Object* type = argv[0]; 
  Scheme_Object* value = argv[1];
  _SES_TYPE_ mysession;

  /* We dont check types of type and value:
   * we are the only ones who use this function. */

  /* Find sessionid for current thread. */
  mysession = get_session_id();

  /* If given _MAX_SES_, warn no more sessions are available */
  if(mysession == _MAX_SES_) {
    scheme_warning("No More Sessions Available");
    return spico_error;
  }

  /* Switch over looptype. 
   * spico_continue: We previously escaped continuation, needs to be continued. 
   * spico_start   : We are starting a new session, start evaluating the
   *                 given scheme-string (which contains the program to evaluate
   * spico_accept  : Input was requested by pico, accept given value.
   *
   * Every one of the cases will save a resulttype to res. */
  if(type == spico_continue) {
    res = _PICO_CONTINUE_(mysession);
  } else if (type == spico_start){
    res = _PICO_DO_(mysession,SCHEME_STR_VAL(value));
  } else if (type == spico_accept) {
    res = _PICO_ACCEPT_(mysession, SCHEME_STR_VAL(value));
  }

  while (res == _DO_LOAD_) {
    /* We will try to read the given file. */
    if (read_file(sessions[mysession]), globalfilebuf != NULL) {
      /* If it succeeds, we empty the message-buffer, and _PICO_LOAD_
       * the program-text. Next we free the globalfilebuf. */
      sessions[mysession].message[0] = '\0';
      res = _PICO_LOAD_(mysession, globalfilebuf);
      free(globalfilebuf);
      globalfilebuf = NULL;
    } else {
      /* If loading of file failed, notify an error has occured */
      sessions[mysession].message[0] = '\0';
      return spico_error;
    }
  }

  switch (res) {
    case _PICO_DONE_:
      {
        /* When pico has finished, the outputobject of the current
         * session will be returned. We also reset the output-pointer 
         * to 0. */
        Scheme_Object* output = scheme_make_string(sessions[mysession].output);
        sessions[mysession].output[0] = '\0';
        return output;
      }
    case _DO_ESCAPE_:
      {
        /* A _DO_ESCAPE_ means that
         * pico wants to give control to other sessions.
         * We use this opportunity to 
         * check here if the break-button was pressed. */
        if(scheme_break_waiting(scheme_current_thread)) {
          sessions[mysession].output[0] = '\0';
          return spico_error;
        } else {
          return spico_wait;
        }
      }
    case _DO_ACCEPT_:
      { /* Notify pico wants input from user */
        return spico_accept;
      }
    case _USR_ERROR_:
      {
        /* When a user wants to print a custom error-message, we print
         * Error: <custommessage> */
        sprintf(sprintfbuf, "Error: %s", sessions[mysession].message);
        scheme_warning(sprintfbuf);
        return spico_error;
      }
    default:
      {
        /* Error section */
        if(res < _PICO_DONE_) {
          /* Fatal */
          scheme_warning("FATAL ERROR: Please Restart Evaluation");
          return spico_error;
        }
        else {
          /* Non fatal */

          /* If there is a marking possible (thus start of errorpos != 0)
           * we also return a pair containing start and stopposition of
           * marking, so we know in scheme we want to mark, and what
           * needs to be marked. */
          scheme_warning(errors[res]);
          if(sessions[mysession].errorpos.start != 0) {
            Scheme_Object* start = 
              scheme_make_integer(sessions[mysession].errorpos.start);
            Scheme_Object* stop =
              scheme_make_integer(sessions[mysession].errorpos.stop);
            sessions[mysession].errorpos.start = 0;
            sessions[mysession].output[0] = '\0';
            return scheme_make_pair(start,stop);
          }
        }
      }
  }

  /* Should not reach. */
  return spico_error;
}

/* Initializing the module, by calling the _PICO_INIT_,
 * creating sessions, loading scheme-symbols to interact with scheme,
 * and adding the pico-eval function to the spico-eval scheme-module. */
Scheme_Object *scheme_initialize(Scheme_Env *env) {
  char *mem;
  Scheme_Env* environment;

  memset(&sessions, 0, sizeof(SessionContext)*_MAX_SES_);

  mem = (char *)malloc(MEMSIZE);
  _PICO_INIT_(mem,MEMSIZE);
  
  
  spico_error = scheme_intern_exact_symbol("spico-error",11);
  spico_continue = scheme_intern_exact_symbol("spico-continue",14);
  spico_accept = scheme_intern_exact_symbol("spico-accept",12);
  spico_wait = scheme_intern_exact_symbol("spico-wait",10);
  spico_start = scheme_intern_exact_symbol("spico-start",11);

  environment = scheme_primitive_module (scheme_intern_symbol ("spico-eval"), env);
  scheme_add_global("pico-eval",scheme_make_prim_w_arity(pico_eval,"pico-eval",2,2),environment); 
  scheme_finish_primitive_module (environment);

   
  return scheme_void;
}

/* When reloading scheme, we just re-initialize scheme. */
Scheme_Object *scheme_reload(Scheme_Env *env) {
  return scheme_initialize(env);
}

Scheme_Object *scheme_module_name() {
  return scheme_intern_symbol ("spico-eval");
}

